@@ -3,14 +3,13 @@
 #' Generate textural metrics using Grey-Level Co-Occurrence Matrices (GLCM). Can be applied to an entire or image or, if a coterminous
 #' raster of segments is provided, GLCM can be calculated for each segment.
 #'
-#' The underlying C++ code for computing GLCMs and their statistics was originally written by Joel Carlson for the
-#' defunct [radiomics](https://github.com/cran/radiomics) library. It has been reused here with permission from the author.
-#'
 #' @param image SpatRaster. A single-band raster layer from which texture is measured
 #' @param segs SpatRaster. A segmented raster. Cell values should be equal to segment numbers. If \code{segs} are not provided,
 #' GLCM will be calculated for the entire image.
 #' @param n_grey integer. Number of grey levels into which the image will be discretized
-#' @param angle integer. Angle at which GLCM will be calculated. Valid inputs are 0, 45, 90, or 135
+#' @param angle integer. Angle at which GLCM will be calculated. Ex.: `c(0,1)`
+#' @param discretize_range numeric. Vector of two values indicating the minimum and maximum input values for discretizing the image.
+#' This can be useful when processing tiles of a larger image, for which you may want to impose a consistent value range.
 #'
 #' @return data.frame
 #'
@@ -36,7 +35,7 @@
 #'
 #' @export
 
-glcm <- function(image, segs = NULL, n_grey = 32, angle = 0){
+glcm <- function(image, segs = NULL, n_grey = 32, angle = c(0,1), discretize_range = NULL){
 
   # Check image
   if(any(dim(image) == 0))     stop("'image' must contain usable values")
@@ -44,16 +43,23 @@ glcm <- function(image, segs = NULL, n_grey = 32, angle = 0){
   if(all(!is.finite(image[]))) stop("'image' must contain usable values")
 
   # Discretize image (this will replace NAs with 0)
-  img_disc <- .discretize_rast(image, n_grey)
+  img_disc <- .discretize_rast(image, n_grey, discretize_range)
 
   # Compute GLCM for whole image
   if(is.null(segs)){
 
+    # Convert to matrix
     img_mat <- terra::as.matrix(img_disc, wide = TRUE)
 
-    out_glcm <- .glcm_stats(.glcm_calc(img_mat, n_grey = n_grey, angle = angle))
+    # 'GLCMTextures' wants the range to be 0 to n_grey-1
+    img_mat <- img_mat - 1
+
+    # Compute GLCM
+    seg_glcm <- GLCMTextures::make_glcm(img_mat, n_levels=n_grey, shift=angle, na.rm=TRUE)
+
+    # Compute GLCM metrics
+    out_glcm <- GLCMTextures::glcm_metrics(seg_glcm)
 
-  # Compute GLCM by segments
   }else{
 
     # Check segments
@@ -74,8 +80,7 @@ glcm <- function(image, segs = NULL, n_grey = 32, angle = 0){
     seg_dfs <- seg_dfs[!names(seg_dfs) %in% c("Inf", "-Inf", "NaN", "NA")]
 
     # Make empty row
-    empty_row <- .glcm_stats(matrix())
-    empty_row[] <- NA
+    empty_row <- GLCMTextures::glcm_metrics(matrix())
 
     # Create worker function
     .glcm_by_seg <- function(seg_df){
@@ -83,98 +88,64 @@ glcm <- function(image, segs = NULL, n_grey = 32, angle = 0){
       seg_df[,"row"] <- seg_df[,"row"] - min(seg_df[,"row"]) + 1
       seg_df[,"col"] <- seg_df[,"col"] - min(seg_df[,"col"]) + 1
 
+      if(all(is.na(seg_df[["val"]]))) return(empty_row)
+
       # NOTE: any space around the segment is filled by 0s at this stage
       seg_mat <- as.matrix(Matrix::sparseMatrix(i = seg_df[,"row"], j =  seg_df[,"col"], x = seg_df[,"val"]))
 
-      if(all(seg_mat == 0)) return(empty_row)
+      # 'sparseMatrix' doesn't allow 0s, so the original discretization will convert values to 1 to n_grey.
+      # 'GLCMTextures', on the other hand, wants the range to be 0 to n_grey-1, and also wants NA values to be NA
+      # Even if this involves extra steps, using 'sparseMatrix' is 10x faster than converting the data.frame
+      # using 'terra'
+      seg_mat[seg_mat == 0] <- NA
+      seg_mat <- seg_mat - 1
 
-      .glcm_stats(.glcm_calc(seg_mat, n_grey = n_grey, angle = angle))
+      # Compute GLCM
+      seg_glcm <- GLCMTextures::make_glcm(seg_mat, n_levels=n_grey, shift=angle, na.rm = T)
+
+      # Compute GLCM metrics
+      seg_metrics <- GLCMTextures::glcm_metrics(seg_glcm)
+
+      return(seg_metrics)
     }
 
     # Apply worker to compute GLCMs
-    out_glcm <- do.call(plyr::rbind.fill, lapply(seg_dfs, .glcm_by_seg))
+    out_glcm <- as.data.frame(do.call(rbind, lapply(seg_dfs, .glcm_by_seg)))
+
+    # Test code for finding which segment is causing a bug
+    # if(false){
+    #   for(i in 1:length(seg_dfs)){
+    #     cat(i,"\n")
+    #     seg_df <- seg_dfs[[i]]
+    #     .glcm_by_seg(seg_df)
+    #   }
+    # }
 
     # Return result
     row.names(out_glcm) <- names(seg_dfs)
-
   }
 
   return(out_glcm)
 }
 
-.glcm_calc <- function(seg_mat, n_grey, angle, d = 1){
 
-  unique_vals <- sort(unique(c(seg_mat)))
-  unique_vals <- unique_vals[unique_vals > 0]
-
-  if(identical(angle, 0)){
-    counts <- glcm0(seg_mat, n_grey = n_grey, d)
-
-  } else if (identical(angle, 45)){
-    counts <- glcm45(seg_mat, n_grey = n_grey, d)
-
-  } else if (identical(angle, 90)){
-    counts <- glcm90(seg_mat, n_grey = n_grey, d)
-
-  } else if (identical(angle, 135)){
-    counts <- glcm135(seg_mat, n_grey = n_grey, d)
-
-  } else {
-    stop("angle must be one of '0', '45', '90', '135'.")
-  }
-
-  # Row 1 and Col 1 hold NA values, remove them
-  counts <- counts[-1, -1, drop = FALSE]
-
-  # Situation where matrix is composed of a single NA
-  if(length(counts) == 0) return(counts)
-
-  counts <- counts[unique_vals, unique_vals, drop = FALSE]
-
-  rownames(counts) <- colnames(counts) <- unique_vals
-
-  # GLCMs should be symmetrical, so the transpose is added
-  counts <- counts + t(counts)
-
-  # Normalize
-  counts <- counts/sum(counts)
-
-  return(counts)
-
-}
-
-.glcm_stats <- function(data){
-
-  #Set up allowed features
-
-  stats_list <- list(
-    "glcm_mean", "glcm_variance", "glcm_autoCorrelation",
-    "glcm_cProminence", "glcm_cShade", "glcm_cTendency",
-    "glcm_contrast", "glcm_correlation", "glcm_differenceEntropy",
-    "glcm_dissimilarity", "glcm_energy", "glcm_entropy",
-    "glcm_homogeneity1", "glcm_homogeneity2", "glcm_IDMN",
-    "glcm_IDN", "glcm_inverseVariance", "glcm_maxProb",
-    "glcm_sumAverage", "glcm_sumEntropy", "glcm_sumVariance"
-  )
-
-  feature_df <- data.frame(lapply(stats_list, function(f) tryCatch(get(f)(data),
-                                                                     error=function(cond) return(NA),
-                                                                     warning=function(cond) return(NA))))
-  colnames(feature_df) <- stats_list
-  return(feature_df)
-}
-
-.discretize_rast <- function(rast, n_grey){
+.discretize_rast <- function(image, n_grey, discretize_range = NULL){
 
   # Get range
-  img_range <- terra::minmax(rast, compute = TRUE)
-  img_min <- img_range[1,]
-  img_max <- img_range[2,]
+  if(is.null(discretize_range)){
+    img_range <- terra::minmax(image, compute = TRUE)
+    img_min <- img_range[1,]
+    img_max <- img_range[2,]
+  }else{
+
+    img_min <- min(discretize_range)
+    img_max <- max(discretize_range)
+  }
 
   if(img_min == img_max){
 
     # This neatly sets NA to 0 and everything else to 1
-    new_values <- as.integer(is.finite(rast[drop = TRUE]))
+    new_values <- as.integer(is.finite(image[drop = TRUE]))
 
   }else{
 
@@ -182,11 +153,8 @@ glcm <- function(image, segs = NULL, n_grey = 32, angle = 0){
 
     levels <- seq(1, n_grey, 1)
 
-    new_values <- as.integer(cut(rast[drop = TRUE], breaks = breaks, labels = levels, include.lowest = TRUE, right = FALSE))
-
-    # Using cut replaced NaN and Inf with NAs
-    new_values[is.na(new_values)] <- 0
+    new_values <- as.integer(cut(image[drop = TRUE], breaks = breaks, labels = levels, include.lowest = TRUE, right = FALSE))
   }
 
-  terra::setValues(rast, new_values)
+  terra::setValues(image, new_values)
 }
